{
  StudioGM(TM) Software(R)
  
  www.studiogm.us
  www.delphicomponents.studiogm.us
  support@delphicomponents.studiogm.us

  History:
    20 Aug 2007: Release v1.3
      * fixed some bugs (procedure ControlMessages)
      * fixed default properties
    25 Oct 2007: Release v1.2
      * fixed some bugs (procedure ControlMessages)
    16 Oct 2007: Release v1.1
      + Moving borders
      + Sizing borders
      + Demo project
      + Package
      + Support Delphi 2005/2006
      * Optimized code refactoring
    08 Aug 2007: Release v1.0
      + Control moving
      + Control sizing
      + Indent.
}
unit sgmMagneticControls;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs;

type
  TMagenticBorders = set of (mbLeft, mbTop, mbRight, mbBottom);

const
  mbAllBorders = [mbLeft, mbTop, mbRight, mbBottom];

type
  TsgmCustomMagneticControl = class(TComponent)
  private
    FControl: TWinControl;
    FIndent: Integer;
    FActiveSizing: Boolean;
    FActiveMoving: Boolean;
    FMovingBorders: TMagenticBorders;
    FSizingBorders: TMagenticBorders;
    procedure SetControl(const Value: TWinControl);
    procedure SetActiveMoving(const Value: Boolean);
    procedure SetActiveSizing(const Value: Boolean);
    procedure SetMovingBorders(const Value: TMagenticBorders);
    procedure SetSizingBorders(const Value: TMagenticBorders);
  protected
    FMode: Integer;
    FOldControlProc: TFarProc;
    FNewControlProc: TFarProc;
    procedure ControlMessages(var Message: TMessage);
    procedure DoMoving(WindowPos: PWindowPos; WorkArea: TRect);
    procedure DoSizing(WindowPos: PWindowPos; WorkArea: TRect);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CreateWndProc;
    procedure DestroyWndProc;
    property Control: TWinControl read FControl write SetControl;
    property ActiveMoving: Boolean read FActiveMoving write SetActiveMoving default True;
    property ActiveSizing: Boolean read FActiveSizing write SetActiveSizing default True;
    property Indent: Integer read FIndent write FIndent;
    property MovingBorders: TMagenticBorders read FMovingBorders
      write SetMovingBorders default mbAllBorders;
    property SizingBorders: TMagenticBorders read FSizingBorders
      write SetSizingBorders default mbAllBorders;
  public
    constructor Create(AOwner: TComponent); override;
  end;

  TsgmMagneticControl = class(TsgmCustomMagneticControl)
  published
    property ActiveMoving;
    property ActiveSizing;
    property Control;
    property Indent;
    property MovingBorders;
    property SizingBorders;
  end;

  TsgmMagneticForm = class(TsgmCustomMagneticControl)
  public
    constructor Create(AOwner: TComponent); override;
  published
    property ActiveMoving;
    property ActiveSizing;
    property Indent;
    property MovingBorders;
    property SizingBorders;
  end;

procedure Register;

implementation

{$R *.RES}

uses Types;

procedure Register;
begin
  RegisterComponents('Samples', [TsgmMagneticControl, TsgmMagneticForm]);
end;

{ TsgmCustomMagneticControl }

constructor TsgmCustomMagneticControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FOldControlProc := nil;
  FNewControlProc := nil;
  FIndent := 30;
  FMode := -1;
  FActiveMoving := True;
  FActiveSizing := True;
  FMovingBorders := mbAllBorders; 
  FSizingBorders := mbAllBorders;
end;

procedure TsgmCustomMagneticControl.CreateWndProc;
begin
  FNewControlProc := Classes.MakeObjectInstance(ControlMessages);
  FOldControlProc := Pointer(GetWindowLong(FControl.Handle, GWL_WNDPROC));
  SetWindowLong(FControl.Handle, GWL_WNDPROC, Longint(FNewControlProc));
end;

procedure TsgmCustomMagneticControl.DestroyWndProc;
begin
  SetWindowLong(FControl.Handle, GWL_WNDPROC, Longint(FOldControlProc));
  Classes.FreeObjectInstance(FNewControlProc);
  FOldControlProc := nil;
  FNewControlProc := nil;
end;

procedure TsgmCustomMagneticControl.ControlMessages(var Message: TMessage);
var
  R: TRect;
begin

  if not Assigned(FControl) then Exit;
  
  if FControl.InheritsFrom(TForm) then
    R := Screen.WorkAreaRect
  else
    R := TWinControl(Owner).ClientRect;

  Message.Result := CallWindowProc(
    FOldControlProc,
    FControl.Handle,
    Message.Msg,
    Message.WParam,
    Message.LParam
  );

  case Message.Msg of
    WM_WINDOWPOSCHANGING: FMode := 1;
    WM_SIZING: FMode := 0;
    WM_MOVING: FMode := 1;
  end;

  if not (csDesigning in FControl.ComponentState) and
    (Message.Msg = WM_WINDOWPOSCHANGING) and
    ((FMode = 0) or (FMode = 1)) then
  begin
    //LockWindowUpdate(FControl.Handle);
    try
      if (FMode = 1) and FActiveMoving then
        DoMoving(PWindowPos(Message.LParam), R);
      if (FMode = 0) and FActiveSizing then
        DoSizing(PWindowPos(Message.LParam), R);
    finally
      //FMode := -1;
      //LockWindowUpdate(0);
    end;
  end;
end;

procedure TsgmCustomMagneticControl.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FControl)
    then
    begin
      if Assigned(FControl) and not (csDestroying in FControl.ComponentState) then
        DestroyWndProc;
      FControl := nil;
    end;
end;

procedure TsgmCustomMagneticControl.SetControl(const Value: TWinControl);
begin
  if FControl <> Value then
  begin
    if Assigned(FControl) and (FOldControlProc <> nil) and (FNewControlProc <> nil)
      then DestroyWndProc;
    FControl := Value;
    if Assigned(FControl)
      then CreateWndProc;
  end;
end;

procedure TsgmCustomMagneticControl.SetActiveMoving(const Value: Boolean);
begin
  if FActiveMoving = Value then Exit;
  FActiveMoving := Value and Assigned(FControl) and Assigned(FNewControlProc);
end;

procedure TsgmCustomMagneticControl.SetActiveSizing(const Value: Boolean);
begin
  if FActiveSizing = Value then Exit;
  FActiveSizing := Value and Assigned(FControl) and Assigned(FNewControlProc);
end;

procedure TsgmCustomMagneticControl.SetMovingBorders(const Value: TMagenticBorders);
begin
  if FMovingBorders = Value then Exit;
  FMovingBorders := Value;
end;

procedure TsgmCustomMagneticControl.SetSizingBorders(const Value: TMagenticBorders);
begin
  if FSizingBorders = Value then Exit;
  FSizingBorders := Value;
end;

procedure TsgmCustomMagneticControl.DoMoving(WindowPos: PWindowPos; WorkArea: TRect);
begin
  with WindowPos^, WorkArea do
  begin
    if (mbLeft in FMovingBorders) and (x < Left+FIndent) then
      x := Left;
    if (mbTop in FMovingBorders) and (y < Top+FIndent) then
      y := Top;
    if (mbRight in FMovingBorders) and (x+cx > Right-FIndent) then
      x := Right - cx;
    if (mbBottom in FMovingBorders) and (y+cy > Bottom-FIndent) then
      y := Bottom - cy;
  end;
end;

procedure TsgmCustomMagneticControl.DoSizing(WindowPos: PWindowPos; WorkArea: TRect);
begin
  with WindowPos^, WorkArea do
  begin
    if (mbLeft in FSizingBorders) and (x < Left+FIndent) then
    begin
      cx := x+cx-Left;
      x := Left;
    end;
    if (mbTop in FSizingBorders) and (y < Top+FIndent) then
    begin
      cy := y+cy-Top;
      y := Top;
    end;
    if (mbRight in FSizingBorders) and (cx+x > Right-FIndent) then
      cx := Right-x;
    if (mbBottom in FSizingBorders) and (cy+y > Bottom-FIndent) then
      cy := Bottom-y;
  end;
end;

{ TsgmMagneticForm }

constructor TsgmMagneticForm.Create(AOwner: TComponent);
var
  FOwner: TComponent;
begin
  inherited Create(AOwner);
  FOwner := AOwner;
  while Assigned(FOwner) do
  begin
    if FOwner.InheritsFrom(TForm) then
    begin
      Control := TForm(FOwner);
      Break;
    end;
    FOwner := FOwner.Owner;
  end;
end;

end.
